package hu.stunts.scr.apps;

import hu.stunts.scr.exceptions.MissingReferenceTrackException;
import hu.stunts.scr.utils.InfoFormatter;
import hu.stunts.scr.utils.TxtParsing;

import java.io.*;
import java.net.URL;
import java.util.*;

public class HierarchyRace<T extends BaseReplay> implements Race {

	private final ClassLoader ccl = Thread.currentThread().getContextClassLoader();
	private URL urlToRacerNames = ccl.getResource("resources/racernames.txt");

	private Class<T> rplClass;
	private String raceTitle;
	private File rootDir;
	private File trkFile;
	private InfoFormatter rFormatter = null;
	private Map<String, String> namesDict = null;
	private Map<String, String> bonusDict = null;
	
	private List<T> activeReplays = new ArrayList<T>();
	private File[] racerDirs;

	private class racerDirsFilter implements FileFilter {
		public boolean accept(File pathname) {
			return (pathname.isDirectory() && pathname.getName() != "extras");
		}
	}
	
	private class ucTRKFilter implements FilenameFilter {
		public boolean accept(File dir, String name) {
			return (name.endsWith(".TRK"));
		}
	}
	
	private class Racer extends AbstractInnerRacer<T> {
		private Map<String, String> penaltyDict = null;
		private String nameKey = null;
		private RplDirectory rplDir;
		
		public Racer (File dir){
			rplDir = new RplDirectory(dir);
			nameKey = dir.getName().toUpperCase();
			//penalty list
			int dictInitSize = 50;
			File penaltyASCII = new File(dir, "penalty.txt");
			if (penaltyASCII.exists())
				penaltyDict = TxtParsing.simpleMakeMapFromPairs(penaltyASCII, dictInitSize);
		}

		public T getBestReplay () {
			String racerName = null;
			T currReplay = null;
			T bestReplay = null;
			if (rplDir.getRplCount() > 0) {
				for (int i = 0; i < rplDir.getRplCount(); i++) {
					currReplay = genericReplay(rplDir.getDirPath() + File.separator + rplDir.getFileName(i), penaltyDict, bonusDict);
					if (currReplay != null && isLegal(currReplay) &&
							(bestReplay == null || (currReplay.compareTo(bestReplay) < 0)))
						bestReplay = currReplay;

				}
			}		
			racerName = namesDict.get(nameKey);
			if (racerName == null)
				racerName = nameKey;
			if (bestReplay != null)
				bestReplay.setRacerName(racerName);
			return bestReplay;
		}

		public T genericReplay (String fileName, Map<String, String>... dictionaries) {
			return genericReplay(rplClass, fileName, dictionaries);
		}
		
		public boolean isLegal (T replay) {
			boolean answer = false;
			try {
				answer = (replay.isValid() &&
						replay.checkTermination());
				if (trkFile != null && trkFile.exists())
					answer = answer && replay.checkTrack(trkFile);
			} catch (MissingReferenceTrackException e) {
				System.out.println("Missing track!");
				e.printStackTrace();
			} catch (IOException e) {
				System.out.println("IO error on track matching!");
				e.printStackTrace();
			}
			return answer;
		} 
	}

	public static <T extends BaseReplay> HierarchyRace<T> create(File root, Class<T> replayClass) {
		return new HierarchyRace<T>(root, replayClass);
	}
	
	protected HierarchyRace (File root, Class<T> replayClass) {
		rplClass = replayClass;
		rootDir = root;
		rFormatter = new InfoFormatter(root);
		if (rootSanityCheck());
		//track file
		String[] trkList = rootDir.list(new ucTRKFilter());
		if (trkList.length > 0) {
			trkFile = new File(rootDir, trkList[0]);
			raceTitle = trkFile.getName().substring(0, trkFile.getName().length() - 4);
			if (trkList.length > 1)
				System.err.println("More than one .TRK on root dir, expect unwanted results.");
		}
		else
			trkFile = null;
		//bonus list
		int dictInitSize = 50;	
		if (rplClass.isAssignableFrom(BonusTableReplay.class)) {
			File bonusASCII = new File(rootDir, "carbonus.txt");
			if (bonusASCII.exists())
				bonusDict = TxtParsing.simpleMakeMapFromPairs(bonusASCII, dictInitSize);
		}
		//names list
		dictInitSize = 100;
		try {
			InputStream namesIS = urlToRacerNames.openStream();
			namesDict = TxtParsing.simpleMakeMapFromPairs(namesIS, dictInitSize, true);
		} catch (IOException e) {
			e.printStackTrace();
		}
		File extraNamesASCII = new File(rootDir, "racernames.txt");
		if (extraNamesASCII.exists()) {
			Map<String, String> extraNamesDict = TxtParsing.simpleMakeMapFromPairs(extraNamesASCII, dictInitSize, true);
			namesDict.putAll(extraNamesDict);
		}

	}

	public void loadActiveReplays () {
		racerDirs = rootDir.listFiles(new racerDirsFilter());
		activeReplays.clear();
		for (File dir : racerDirs) {
			Racer racer = new Racer(dir);
			T activeReplay = racer.getBestReplay();
			if (activeReplay != null)
				activeReplays.add(activeReplay);
		}
		Collections.sort(activeReplays);
	}
	
	private class ScoreboardGenerator extends AbstractScoreboardGenerator {
		public ScoreboardGenerator() {
			super(activeReplays, rFormatter);
		}
	}
	
	public String[][] generateHIGStrings (boolean useCorrectedTimes) {
		ScoreboardGenerator sg = new ScoreboardGenerator();
		return sg.generateHIGStrings(useCorrectedTimes);
	}
			
	public void generateHIG (File dest, boolean correctTimes) {
		ScoreboardGenerator sg = new ScoreboardGenerator();
		sg.writeHIG(dest, correctTimes);
	}
	
	public String getRaceTitle() {
		return raceTitle;
	}
	
	public File getRootDir() {
		return rootDir;
	}
	
	public void printTest () {
		/*for (T replay : activeReplays) {
			System.out.println(replay.getRacer()+" : "+replay.toSeconds(replay.computeTime()));
		}*/
		String[][] higData = generateHIGStrings(false); 
		for (String[] row : higData) {
			for (String field : row)
				System.out.print(field + "    ");
			System.out.println("");
		}	
	}
	
	
	protected boolean rootSanityCheck () {//TODO in a future release
		return true;
	}

}
